﻿#include "pch.h"
#include "PaginationExperiment1Renderer.h"
#include "Common/DirectXHelper.h"

using namespace PaginationExperiment1;

using namespace D2D1;

PaginationExperiment1Renderer::PaginationExperiment1Renderer(const std::shared_ptr<DeviceResources>& deviceResources) :
    m_currentScroll(0),
    m_maximumScroll(0),
	m_deviceResources(deviceResources)
{
	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

	CreateDeviceDependentResources();
}

void PaginationExperiment1Renderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}
void PaginationExperiment1Renderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void PaginationExperiment1Renderer::CreateWindowSizeDependentResources()
{
    PaginateChapter();
}

void PaginationExperiment1Renderer::SetAliceText(std::vector<std::wstring> lines)
{
    m_aliceParagraphGenerator.SetText(lines);
    PaginateChapter();
}

void PaginationExperiment1Renderer::PaginateChapter()
{
    if (m_aliceParagraphGenerator.ParagraphCount() == 0)
        return;

    Windows::Foundation::Size windowSize = m_deviceResources->GetLogicalSize();

    if (windowSize.IsEmpty)
        return;

    // Set some constants and calculate page and text sizes
    float pageHeight = windowSize.Height;
    float topMargin = 25;
    float bottomMargin = 25;
    float textHeight = pageHeight - topMargin - bottomMargin;

    float pageWidth = 400;
    float leftMargin = 25;
    float rightMargin = 25;
    float textWidth = pageWidth - leftMargin - rightMargin;

    // Set the width of all the paragraphs; also generates formatted heights
    m_aliceParagraphGenerator.SetWidth(textWidth);

    // Prepare for assembling paragraphs into pages
    std::vector<Paragraph> paragraphs = 
                    m_aliceParagraphGenerator.GetParagraphs();
    m_pages.clear();

    // Prepare for the first page
    float x = leftMargin;
    float y = topMargin;
    float maxY = pageHeight - bottomMargin;
    Page page;
    page.Size = SizeF(pageWidth, pageHeight);   // uniform page size

    // Loop through all the paragraphs
    for (Paragraph& paragraph : paragraphs)
    {
        // Check if the paragraph will make the page too long
        if (y + paragraph.TextHeight > maxY)
        {
            // If so, save the page and begin a new page
            m_pages.push_back(page);
            x = leftMargin;
            y = topMargin;
            page.Paragraphs.clear();
        }

        // Determine where the paragraph should appear and save it
        paragraph.Origin = Point2F(x, y);
        page.Paragraphs.push_back(paragraph);

        // Bump up the y coordinate for the next paragraph
        y += paragraph.TextHeight + paragraph.SpaceAfter;
    }

    // Save the last page as well
    m_pages.push_back(page);

    // Calculate the maximum scroll
    m_maximumScroll = m_pages.size() * pageWidth - windowSize.Width;
    ScrollHorizontalDelta(0);
}

void PaginationExperiment1Renderer::ScrollHorizontalDelta(float deltaX)
{
    m_currentScroll = max(0, min(m_maximumScroll, m_currentScroll - deltaX));
}

void PaginationExperiment1Renderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void PaginationExperiment1Renderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::White));

    Matrix3x2F scrollTransform = Matrix3x2F::Translation(-m_currentScroll, 0);
    context->SetTransform(scrollTransform * 
                          m_deviceResources->GetOrientationTransform2D());

    for (const Page& page : m_pages)
    {
        for (const Paragraph& paragraph : page.Paragraphs)
        {
            context->DrawTextLayout(paragraph.Origin, 
                                    paragraph.TextLayout.Get(), 
                                    m_blackBrush.Get());
        }

        // Set transform for next column
        D2D1_MATRIX_3X2_F xform;
        context->GetTransform(&xform);
        xform = Matrix3x2F::Translation(page.Size.width, 0) * xform;
        context->SetTransform(&xform);
    }

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

