#include "pch.h"
#include "SplittableTextLayout.h"

using namespace Microsoft::WRL;

HRESULT SplittableTextLayout::Create(_In_ const WCHAR * string,
                                     UINT32  stringLength,
                                     IDWriteTextFormat * textFormat,
                                     FLOAT  maxWidth,
                                     FLOAT  maxHeight,
                                     _Outptr_ SplittableTextLayout ** splittableTextLayout)
{
    ComPtr<IDWriteFactory2> dwriteFactory;

    HRESULT hr = DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED,
                                    __uuidof(IDWriteFactory2),
                                    &dwriteFactory);

    if (FAILED(hr))
        return hr;

    ComPtr<IDWriteTextLayout> dwriteTextLayout;
    
    hr = dwriteFactory->CreateTextLayout(string, 
                                         stringLength, 
                                         textFormat, 
                                         maxWidth, 
                                         maxHeight, 
                                         &dwriteTextLayout);

    if (FAILED(hr))
        return hr;

    SplittableTextLayout* pSplittableTextLayout = new SplittableTextLayout();

    pSplittableTextLayout->m_dwriteFactory = dwriteFactory;
    pSplittableTextLayout->m_dwriteTextFormat = textFormat;
    pSplittableTextLayout->m_string = string;
    pSplittableTextLayout->m_stringLength = stringLength;
    pSplittableTextLayout->m_maxWidth = maxWidth;
    pSplittableTextLayout->m_maxHeight = maxHeight;

    dwriteTextLayout.As(&pSplittableTextLayout->m_dwriteTextLayout);

    *splittableTextLayout = pSplittableTextLayout;
    return S_OK;
}

SplittableTextLayout::SplittableTextLayout() : m_refCount(1)
{
}

// IUnknown methods
ULONG STDMETHODCALLTYPE SplittableTextLayout::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE SplittableTextLayout::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE SplittableTextLayout::QueryInterface(_In_ REFIID riid, _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(IDWriteTextLayout1))
    {
        *ppOutput = static_cast<IDWriteTextLayout1*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWriteTextLayout))
    {
        *ppOutput = static_cast<IDWriteTextLayout*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWriteTextFormat))
    {
        *ppOutput = static_cast<IDWriteTextFormat*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

HRESULT SplittableTextLayout::Split(UINT32 stringOffset, 
                                    _Outptr_ SplittableTextLayout ** splittableTextLayout1, 
                                    _Outptr_ SplittableTextLayout ** splittableTextLayout2)
{
    if (stringOffset < 1 || stringOffset >= m_stringLength)
        return E_INVALIDARG;

    std::wstring str1 = std::wstring(m_string).substr(0, stringOffset);
    std::wstring str2 = std::wstring(m_string).substr(stringOffset, m_stringLength - stringOffset);

    SplittableTextLayout* textLayout1;
    HRESULT hr = SplittableTextLayout::Create(str1.data(), 
                                              str1.length(), 
                                              m_dwriteTextFormat.Get(), 
                                              m_maxWidth, 
                                              m_maxHeight, 
                                              &textLayout1);
    if (FAILED(hr))
        return hr;

    SplittableTextLayout* textLayout2;
    hr = SplittableTextLayout::Create(str2.data(), 
                                      str2.length(), 
                                      m_dwriteTextFormat.Get(), 
                                      m_maxWidth, 
                                      m_maxHeight, 
                                      &textLayout2);
    if (FAILED(hr))
        return hr;

    *splittableTextLayout1 = textLayout1;
    *splittableTextLayout2 = textLayout2;

    return S_OK;
}