﻿#include "pch.h"
#include "PaginationExperiment3Renderer.h"
#include "Common/DirectXHelper.h"

using namespace PaginationExperiment3;

using namespace D2D1;
using namespace Microsoft::WRL;

PaginationExperiment3Renderer::PaginationExperiment3Renderer(const std::shared_ptr<DeviceResources>& deviceResources) :
    m_currentScroll(0),
    m_maximumScroll(0),
	m_deviceResources(deviceResources)
{
	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

	CreateDeviceDependentResources();
}

void PaginationExperiment3Renderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}
void PaginationExperiment3Renderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void PaginationExperiment3Renderer::CreateWindowSizeDependentResources()
{
    PaginateChapter();
}

void PaginationExperiment3Renderer::SetAliceText(std::vector<std::wstring> lines)
{
    m_aliceParagraphGenerator.SetText(lines);
    PaginateChapter();
}

void PaginationExperiment3Renderer::PaginateChapter()
{
    if (m_aliceParagraphGenerator.ParagraphCount() == 0)
        return;

    Windows::Foundation::Size windowSize = m_deviceResources->GetLogicalSize();

    if (windowSize.IsEmpty)
        return;

    // Set some constants and calculate page and text sizes
    float pageHeight = windowSize.Height;
    float topMargin = 25;
    float bottomMargin = 25;
    float textHeight = pageHeight - topMargin - bottomMargin;

    float pageWidth = 400;
    float leftMargin = 25;
    float rightMargin = 25;
    float textWidth = pageWidth - leftMargin - rightMargin;

    // Prepare assembling pages
    std::vector<Paragraph> paragraphs = m_aliceParagraphGenerator.GetParagraphs();
    m_pages.clear();

    // Prepare for the first page
    float x = leftMargin;
    float y = topMargin;
    float maxY = pageHeight - bottomMargin;
    Page page;
    page.Size = SizeF(pageWidth, pageHeight);   // uniform page size

    // Loop through all the paragraphs
    for (size_t i = 0; i < paragraphs.size(); i++)
    {
        Paragraph paragraph = paragraphs.at(i);

        // Set paragraph height and width
        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxWidth(textWidth)
            );

        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxHeight(FloatMax())
            );

        // Add paragraph to page (or possibly pages)
        bool isSplit = false;

        do
        {
            while (AddParagraphToPage(page, paragraph, x, y, maxY, isSplit))
            {
                m_pages.push_back(page);
                x = leftMargin;
                y = topMargin;
                page.Paragraphs.clear();
            }
        } while (isSplit);
    }

    m_pages.push_back(page);

    // Calculate the maximum scroll
    m_maximumScroll = m_pages.size() * pageWidth - windowSize.Width;
    ScrollHorizontalDelta(0);
}

bool PaginationExperiment3Renderer::AddParagraphToPage(Page& page, 
                                                       Paragraph& paragraph, 
                                                       float x, 
                                                       float& y, 
                                                       float maxY, 
                                                       bool& isSplit)
{
    // We already know where the paragraph is going
    paragraph.Origin = Point2F(x, y);

    // Obtain the height of the paragraph
    DWRITE_TEXT_METRICS textMetrics;

    DX::ThrowIfFailed(
        paragraph.TextLayout->GetMetrics(&textMetrics)
        );

    float paragraphHeight = textMetrics.height;

    // Check if it overruns the maximum height
    if (y + paragraphHeight > maxY)
    {
        // Paragraph is too long, check for a split
        UINT32 lineCount = textMetrics.lineCount;

        std::vector<DWRITE_LINE_METRICS> lineMetrics(lineCount);

        DX::ThrowIfFailed(
            paragraph.TextLayout->GetLineMetrics(lineMetrics.data(), 
                                                 lineMetrics.size(), 
                                                 &lineCount)
            );

        // Accumulate the number of characters in the lines that fit
        int characterCount = 0;

        for (DWRITE_LINE_METRICS lineMetric : lineMetrics)
        {
            if (y + lineMetric.height > maxY)
            {
                break;
            }
            characterCount += lineMetric.length;
            y += lineMetric.height;
        }

        // At this point, characterCount might be zero, but is less than the paragraph
        if (characterCount != 0)
        {
            ComPtr<SplittableTextLayout> textLayout1;
            ComPtr<SplittableTextLayout> textLayout2;

            DX::ThrowIfFailed(
                paragraph.TextLayout->Split(characterCount, &textLayout1, &textLayout2)
                );

            // This is the displayable first half of the paragraph
            paragraph.TextLayout = textLayout1;
            page.Paragraphs.push_back(paragraph);

            // This is the second half of the paragraph
            paragraph.TextLayout = textLayout2;
        }

        isSplit = characterCount != 0;
        return true;
    }

    y += paragraphHeight + paragraph.SpaceAfter;
    page.Paragraphs.push_back(paragraph);
    isSplit = false;
    return false;
}

void PaginationExperiment3Renderer::ScrollHorizontalDelta(float deltaX)
{
    m_currentScroll = max(0, min(m_maximumScroll, m_currentScroll - deltaX));
}

void PaginationExperiment3Renderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void PaginationExperiment3Renderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::White));

    Matrix3x2F scrollTransform = Matrix3x2F::Translation(-m_currentScroll, 0);
    context->SetTransform(scrollTransform * m_deviceResources->GetOrientationTransform2D());

    for (const Page& page : m_pages)
    {
        for (const Paragraph& paragraph : page.Paragraphs)
        {
            context->DrawTextLayout(paragraph.Origin, 
                                    paragraph.TextLayout.Get(), 
                                    m_blackBrush.Get());
        }

        // Set transform for next column
        D2D1_MATRIX_3X2_F xform;
        context->GetTransform(&xform);
        xform = Matrix3x2F::Translation(page.Size.width, 0) * xform;
        context->SetTransform(&xform);
    }

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

