﻿#include "pch.h"
#include "PaginationExperiment3Main.h"
#include "Common\DirectXHelper.h"

using namespace PaginationExperiment3;
using namespace Windows::Foundation;
using namespace Windows::System::Threading;
using namespace Concurrency;

// Loads and initializes application assets when the application is loaded.
PaginationExperiment3Main::PaginationExperiment3Main(const std::shared_ptr<DeviceResources>& deviceResources) :
	m_deviceResources(deviceResources)
{
	// Register to be notified if the Device is lost or recreated
	m_deviceResources->RegisterDeviceNotify(this);

	// Content initialization.
	m_PaginationExperiment3Renderer = std::unique_ptr<PaginationExperiment3Renderer>(new PaginationExperiment3Renderer(m_deviceResources));

	// TODO: Change the timer settings if you want something other than the default variable timestep mode.
	// e.g. for 60 FPS fixed timestep update logic, call:
	/*
	m_timer.SetFixedTimeStep(true);
	m_timer.SetTargetElapsedSeconds(1.0 / 60);
	*/
}

PaginationExperiment3Main::~PaginationExperiment3Main()
{
	// Deregister device notification
	m_deviceResources->RegisterDeviceNotify(nullptr);
}

// Updates application state when the window size changes (e.g. device orientation change)
void PaginationExperiment3Main::CreateWindowSizeDependentResources() 
{
	// Size-dependent initialization of your app's content.
    m_PaginationExperiment3Renderer->CreateWindowSizeDependentResources();
}

void PaginationExperiment3Main::StartRenderLoop()
{
	// If the animation render loop is already running then do not start another thread.
	if (m_renderLoopWorker != nullptr && m_renderLoopWorker->Status == AsyncStatus::Started)
	{
		return;
	}

	// Create a task that will be run on a background thread.
	auto workItemHandler = ref new WorkItemHandler([this](IAsyncAction ^ action)
	{
		// Calculate the updated frame and render once per vertical blanking interval.
		while (action->Status == AsyncStatus::Started)
		{
			critical_section::scoped_lock lock(m_criticalSection);
			Update();
			if (Render())
			{
				m_deviceResources->Present();
			}
		}
	});

	// Run task on a dedicated high priority background thread.
	m_renderLoopWorker = ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);
}

void PaginationExperiment3Main::StopRenderLoop()
{
	m_renderLoopWorker->Cancel();
}

// Updates the application state once per frame.
void PaginationExperiment3Main::Update() 
{
	ProcessInput();

	// Update scene objects.
	m_timer.Tick([&]()
	{
		// Content update functions.
		m_PaginationExperiment3Renderer->Update(m_timer);
	});
}

// Process all input from the user before updating game state
void PaginationExperiment3Main::ProcessInput()
{
	// TODO: Add per frame input handling here.
}

// Renders the current frame according to the current application state.
// Returns true if the frame was rendered and is ready to be displayed.
bool PaginationExperiment3Main::Render() 
{
	// Don't try to render anything before the first Update.
	if (m_timer.GetFrameCount() == 0)
	{
		return false;
	}

	auto context = m_deviceResources->GetD3DDeviceContext();

	// Reset the viewport to target the whole screen.
	auto viewport = m_deviceResources->GetScreenViewport();
	context->RSSetViewports(1, &viewport);

	// Reset render targets to the screen.
	ID3D11RenderTargetView *const targets[1] = { m_deviceResources->GetBackBufferRenderTargetView() };
	context->OMSetRenderTargets(1, targets, m_deviceResources->GetDepthStencilView());

	// Clear the back buffer and depth stencil view.
	context->ClearRenderTargetView(m_deviceResources->GetBackBufferRenderTargetView(), DirectX::Colors::CornflowerBlue);
	context->ClearDepthStencilView(m_deviceResources->GetDepthStencilView(), D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);

	// Render the scene objects.
	m_PaginationExperiment3Renderer->Render();

	return true;
}

// Notifies renderers that device resources need to be released.
void PaginationExperiment3Main::OnDeviceLost()
{
	m_PaginationExperiment3Renderer->ReleaseDeviceDependentResources();
}

// Notifies renderers that device resources may now be recreated.
void PaginationExperiment3Main::OnDeviceRestored()
{
	m_PaginationExperiment3Renderer->CreateDeviceDependentResources();
	CreateWindowSizeDependentResources();
}

// Methods called from DirectXPage
void PaginationExperiment3Main::SetAliceText(std::vector<std::wstring> lines)
{
    m_PaginationExperiment3Renderer->SetAliceText(lines);
}

void PaginationExperiment3Main::ScrollHorizontalDelta(float deltaX)
{
    m_PaginationExperiment3Renderer->ScrollHorizontalDelta(deltaX);
}

