﻿#include "pch.h"
#include "PaginationExperiment5Renderer.h"
#include "Common/DirectXHelper.h"

using namespace PaginationExperiment5;

using namespace D2D1;
using namespace Microsoft::WRL;

PaginationExperiment5Renderer::PaginationExperiment5Renderer(const std::shared_ptr<DeviceResources>& deviceResources) :
    m_currentScroll(0),
    m_maximumScroll(0),
	m_deviceResources(deviceResources)
{
	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

	CreateDeviceDependentResources();
}

void PaginationExperiment5Renderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}
void PaginationExperiment5Renderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void PaginationExperiment5Renderer::CreateWindowSizeDependentResources()
{
    PaginateChapter();
}

void PaginationExperiment5Renderer::SetAliceText(std::vector<std::wstring> lines)
{
    m_aliceParagraphGenerator.SetText(lines);
    PaginateChapter();
}

void PaginationExperiment5Renderer::PaginateChapter()
{
    if (m_aliceParagraphGenerator.ParagraphCount() == 0)
        return;

    Windows::Foundation::Size windowSize = m_deviceResources->GetLogicalSize();

    if (windowSize.IsEmpty)
        return;

    // Set some constants and calculate page and text sizes
    float pageHeight = windowSize.Height;
    float topMargin = 25;
    float bottomMargin = 25;
    float textHeight = pageHeight - topMargin - bottomMargin;

    float pageWidth = 400;
    float leftMargin = 25;
    float rightMargin = 25;
    float textWidth = pageWidth - leftMargin - rightMargin;

    // Prepare assembling pages
    std::vector<Paragraph> paragraphs = m_aliceParagraphGenerator.GetParagraphs();
    m_pages.clear();

    // Prepare for the first page
    float x = leftMargin;
    float y = topMargin;
    float maxY = pageHeight - bottomMargin;
    Page page;
    page.Size = SizeF(pageWidth, pageHeight);   // uniform page size

    // Loop through all the paragraphs
    for (size_t i = 0; i < paragraphs.size(); i++)
    {
        Paragraph paragraph = paragraphs.at(i);

        // Set paragraph height and width
        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxWidth(textWidth)
            );

        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxHeight(FloatMax())
            );

        // Add paragraph to page (or possibly pages)
        bool isSplit = false;

        do
        {
            while (AddParagraphToPage(page, paragraph, x, y, maxY, isSplit))
            {
                // End of page so adjust line spacing & add page to collection
                MakePageUniformHeight(page, maxY);
                m_pages.push_back(page);

                // Begin a new page
                x = leftMargin;
                y = topMargin;
                page.Paragraphs.clear();
            }
        } while (isSplit);
    }

    // Don't adjust line spacing on this last page!
    m_pages.push_back(page);

    // Calculate the maximum scroll
    m_maximumScroll = m_pages.size() * pageWidth - windowSize.Width;
    ScrollHorizontalDelta(0);
}

bool PaginationExperiment5Renderer::AddParagraphToPage(Page& page, 
                                                       Paragraph& paragraph, 
                                                       float x, 
                                                       float& y, 
                                                       float maxY, 
                                                       bool& isSplit)
{
    // We already know where the paragraph is going
    paragraph.Origin = Point2F(x, y);

    // Obtain the height of the paragraph
    DWRITE_TEXT_METRICS textMetrics;

    DX::ThrowIfFailed(
        paragraph.TextLayout->GetMetrics(&textMetrics)
        );

    // Save information in Paragraph object
    paragraph.LineCount = textMetrics.lineCount;
    paragraph.Height = textMetrics.height;

    // Check if it overruns the maximum height
    if (y + paragraph.Height > maxY)
    {
        // Paragraph is too long, check for a split
        UINT32 lineCount = textMetrics.lineCount;

        // We already know the paragraph is too long,
        //  so if it consists of only one line, it won't fit.
        // If it has two lines, only one line will fit,
        //  and that creates both a widow and orphan.
        // If it has three lines, either an orphan or 
        //  widow will result.
        // Hence...
        if (paragraph.LineCount <= 3)
        {
            isSplit = false;
            return true;            // end of page
        }

        // Get the line metrics
        std::vector<DWRITE_LINE_METRICS> lineMetrics(lineCount);

        DX::ThrowIfFailed(
            paragraph.TextLayout->GetLineMetrics(lineMetrics.data(), 
                                                 lineMetrics.size(), 
                                                 &lineCount)
            );

        // Accumulate the number of lines and character count that fit
        int fittableLineCount = 0;
        float fittableHeight = 0;
        int characterCount = 0;

        for (DWRITE_LINE_METRICS lineMetric : lineMetrics)
        {
            // Check for next line that exceeds page height
            if (y + lineMetric.height > maxY)
            {
                break;
            }

            fittableLineCount++;
            fittableHeight += lineMetric.height;
            characterCount += lineMetric.length;
            y += lineMetric.height;

            // Check for maximum lines before a widow is made
            if (fittableLineCount + 2 == lineCount)
            {
                break;
            }
        }
        
        // Check for 0 or 1 line (an orphan) that fits on the page
        if (fittableLineCount <= 1)
        {
            isSplit = false;
            return true;            // end of page
        }
        
        // Ready for a valid split!
        ComPtr<SplittableTextLayout> textLayout1;
        ComPtr<SplittableTextLayout> textLayout2;

        DX::ThrowIfFailed(
            paragraph.TextLayout->Split(characterCount, &textLayout1, &textLayout2)
            );

        // This is the displayable first half of the paragraph.
        // Create a new Paragraph with this new layout.
        Paragraph newParagraph = paragraph;
        newParagraph.TextLayout = textLayout1;
        newParagraph.Height = fittableHeight;
        newParagraph.SpaceAfter = 0;            // because at end of page
        newParagraph.LineCount = fittableLineCount;
        page.Paragraphs.push_back(newParagraph);
        page.TextBottom = y;

        // This is the second half of the paragraph,
        //  to be handled in next call to method.
        paragraph.TextLayout = textLayout2;

        isSplit = true;
        return true;            // end of page
    }

    y += paragraph.Height + paragraph.SpaceAfter;

    // Create a new Paragraph with a cloned SplittableTextLayout
    Paragraph newParagraph = paragraph;

    DX::ThrowIfFailed(
        paragraph.TextLayout->Clone(&newParagraph.TextLayout)
        );

    // Add it to the page and update page height
    page.Paragraphs.push_back(newParagraph);
    page.TextBottom = y - paragraph.SpaceAfter;
    isSplit = false;
    return false;               // not end of page
}

void PaginationExperiment5Renderer::MakePageUniformHeight(Page& page, 
                                                          float desiredHeight)
{
    int totalLineCount = 0;

    for (Paragraph& paragraph : page.Paragraphs)
    {
        if (paragraph.CanAdjustLineSpacing)
        {
            totalLineCount += paragraph.LineCount;
        }
    }

    float extraLineSpacing = (desiredHeight - page.TextBottom) / totalLineCount;
    D2D1_POINT_2F origin = page.Paragraphs[0].Origin;

    for (Paragraph& paragraph : page.Paragraphs)
    {
        // Set new origin
        paragraph.Origin = origin;

        if (paragraph.CanAdjustLineSpacing)
        {
            // Calculate and set new line spacing
            float lineSpacing = paragraph.Height / paragraph.LineCount;
            lineSpacing += extraLineSpacing;

            DX::ThrowIfFailed(
                paragraph.TextLayout->SetLineSpacing
                                (DWRITE_LINE_SPACING_METHOD_UNIFORM,
                                 lineSpacing, 0.8f * lineSpacing)
                );

            // Set new paragraph height
            paragraph.Height = lineSpacing * paragraph.LineCount;
        }

        // Calculate origin of next paragraph
        origin.y += paragraph.Height + paragraph.SpaceAfter;
    }
}

void PaginationExperiment5Renderer::ScrollHorizontalDelta(float deltaX)
{
    m_currentScroll = max(0, min(m_maximumScroll, m_currentScroll - deltaX));
}

void PaginationExperiment5Renderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void PaginationExperiment5Renderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::White));

    Matrix3x2F scrollTransform = Matrix3x2F::Translation(-m_currentScroll, 0);
    context->SetTransform(scrollTransform * m_deviceResources->GetOrientationTransform2D());

    for (const Page& page : m_pages)
    {
        for (const Paragraph& paragraph : page.Paragraphs)
        {
            context->DrawTextLayout(paragraph.Origin, 
                                    paragraph.TextLayout.Get(), 
                                    m_blackBrush.Get());
        }

        // Set transform for next column
        D2D1_MATRIX_3X2_F xform;
        context->GetTransform(&xform);
        xform = Matrix3x2F::Translation(page.Size.width, 0) * xform;
        context->SetTransform(&xform);
    }

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

