#include "pch.h"
#include "SplittableTextLayout.h"

using namespace Microsoft::WRL;

HRESULT SplittableTextLayout::Create(_In_ const WCHAR * string,
                                     UINT32  stringLength,
                                     IDWriteTextFormat * textFormat,
                                     FLOAT  maxWidth,
                                     FLOAT  maxHeight,
                                     _Outptr_ SplittableTextLayout ** splittableTextLayout)
{
    ComPtr<IDWriteFactory2> dwriteFactory;

    HRESULT hr = DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED,
                                    __uuidof(IDWriteFactory2),
                                    &dwriteFactory);

    if (FAILED(hr))
        return hr;

    ComPtr<IDWriteTextLayout> dwriteTextLayout;
    
    hr = dwriteFactory->CreateTextLayout(string, 
                                         stringLength, 
                                         textFormat, 
                                         maxWidth, 
                                         maxHeight, 
                                         &dwriteTextLayout);

    if (FAILED(hr))
        return hr;

    SplittableTextLayout* pSplittableTextLayout = new SplittableTextLayout();

    pSplittableTextLayout->m_dwriteFactory = dwriteFactory;
    pSplittableTextLayout->m_dwriteTextFormat = textFormat;
    pSplittableTextLayout->m_string = string;
    pSplittableTextLayout->m_stringLength = stringLength;
    pSplittableTextLayout->m_maxWidth = maxWidth;
    pSplittableTextLayout->m_maxHeight = maxHeight;

    dwriteTextLayout.As(&pSplittableTextLayout->m_dwriteTextLayout);

    *splittableTextLayout = pSplittableTextLayout;
    return S_OK;
}

SplittableTextLayout::SplittableTextLayout() : m_refCount(1)
{
}

// IUnknown methods
ULONG STDMETHODCALLTYPE SplittableTextLayout::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE SplittableTextLayout::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE SplittableTextLayout::QueryInterface(_In_ REFIID riid, _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(IDWriteTextLayout1))
    {
        *ppOutput = static_cast<IDWriteTextLayout1*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWriteTextLayout))
    {
        *ppOutput = static_cast<IDWriteTextLayout*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWriteTextFormat))
    {
        *ppOutput = static_cast<IDWriteTextFormat*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

HRESULT SplittableTextLayout::Clone(_Outptr_ SplittableTextLayout ** splittableTextLayout)
{
    return InternalClone(0, m_stringLength, splittableTextLayout, false);
}

HRESULT SplittableTextLayout::InternalClone(int stringOffset, int stringLength,
                                            _Outptr_ SplittableTextLayout ** splittableTextLayout,
                                            bool adjustText)
{
    std::wstring str = std::wstring(m_string).substr(stringOffset, stringLength);

    if (adjustText)
    {
        // Add some space plus non-printable character so last line 
        //  of first paragraph segment has proper alignment
        if (this->GetTextAlignment() == DWRITE_TEXT_ALIGNMENT_JUSTIFIED)
        {
            float fontSize = this->GetFontSize();
            float maxWidth = this->GetMaxWidth();

            // Assume space has width of 5% of em height.
            //  (Very low, but the Parchment font is 9%)
            int numSpaces = (int) (maxWidth / (fontSize * 0.05));

            str += std::wstring(numSpaces, ' ') + L"\x0001";
        }
    }

    SplittableTextLayout* textLayout;
    HRESULT hr = SplittableTextLayout::Create(str.data(),
                                              str.length(),
                                              m_dwriteTextFormat.Get(),
                                              m_maxWidth,
                                              m_maxHeight,
                                              &textLayout);
    if (FAILED(hr))
        return hr;

    // Transfer settings from this instance to the two split text layouts
    DWRITE_FLOW_DIRECTION flowDirection = this->GetFlowDirection();
    textLayout->SetFlowDirection(flowDirection);

    float tabStop = this->GetIncrementalTabStop();
    textLayout->SetIncrementalTabStop(tabStop);

    DWRITE_LINE_SPACING_METHOD lineSpacingMethod;
    float lineSpacing, baseline;
    if (FAILED(
        hr = this->GetLineSpacing(&lineSpacingMethod, &lineSpacing, &baseline))
        ) return hr;
    if (FAILED(
        hr = textLayout->SetLineSpacing(lineSpacingMethod, lineSpacing, baseline))
        ) return hr;

    DWRITE_PARAGRAPH_ALIGNMENT paragraphAlignment = this->GetParagraphAlignment();
    textLayout->SetParagraphAlignment(paragraphAlignment);

    DWRITE_READING_DIRECTION readingDirection = this->GetReadingDirection();
    textLayout->SetReadingDirection(readingDirection);

    DWRITE_TEXT_ALIGNMENT textAlignment = this->GetTextAlignment();
    textLayout->SetTextAlignment(textAlignment);

    DWRITE_TRIMMING trimmingOptions;
    ComPtr<IDWriteInlineObject> trimmingSign;
    if (FAILED(
        hr = this->GetTrimming(&trimmingOptions, &trimmingSign))
        ) return hr;
    if (FAILED(
        hr = textLayout->SetTrimming(&trimmingOptions, trimmingSign.Get()))
        ) return hr;

    DWRITE_WORD_WRAPPING wordWrapping = this->GetWordWrapping();
    textLayout->SetWordWrapping(wordWrapping);

    for (PropertySetting propertySetting : m_propertySettings)
    {
        DWRITE_TEXT_RANGE textRange = propertySetting.TextRange;
        int startRange = textRange.startPosition;
        int endRange = textRange.startPosition + textRange.length;

        startRange = max(0, startRange - stringOffset);
        endRange = min(max(0, endRange - stringOffset), stringLength);

        if (startRange >= 0 && endRange > startRange)
        {
            textRange.startPosition = startRange;
            textRange.length = endRange - startRange;

            if (FAILED(hr = TransferPropertySetting(textLayout,
                                                    propertySetting,
                                                    textRange)))
                return hr;
        }
    }

    *splittableTextLayout = textLayout;
    return S_OK;
}

HRESULT SplittableTextLayout::Split(UINT32 stringOffset, 
                                    _Outptr_ SplittableTextLayout ** splittableTextLayout1, 
                                    _Outptr_ SplittableTextLayout ** splittableTextLayout2)
{
    if (stringOffset < 1 || stringOffset >= m_stringLength)
        return E_INVALIDARG;

    HRESULT hr = InternalClone(0, stringOffset, splittableTextLayout1, true);

    if (FAILED(hr))
        return hr;

    return InternalClone(stringOffset, m_stringLength - stringOffset, splittableTextLayout2, false);
}

void SplittableTextLayout::SaveInterfaceType(PropertyType propertyType, 
                                             IUnknown * pIUnknown, 
                                             DWRITE_TEXT_RANGE textRange)
{
    PropertySetting propertySetting;
    propertySetting.PropertyType = propertyType;
    propertySetting.IUnknownPointer = pIUnknown;
    propertySetting.TextRange = textRange;

    m_propertySettings.push_back(propertySetting);
}

void SplittableTextLayout::SaveStringType(PropertyType propertyType, 
                                          const WCHAR * pString, 
                                          DWRITE_TEXT_RANGE textRange)
{
    PropertySetting propertySetting;
    propertySetting.PropertyType = propertyType;
    propertySetting.StringPointer = pString;
    propertySetting.TextRange = textRange;

    m_propertySettings.push_back(propertySetting);
}

void SplittableTextLayout::SaveFloatType(PropertyType propertyType, 
                                         float val0, float val1, float val2, 
                                         DWRITE_TEXT_RANGE textRange)
{
    PropertySetting propertySetting;
    propertySetting.PropertyType = propertyType;
    propertySetting.FloatValue[0] = val0;
    propertySetting.FloatValue[1] = val1;
    propertySetting.FloatValue[2] = val2;
    propertySetting.TextRange = textRange;

    m_propertySettings.push_back(propertySetting);
}

void SplittableTextLayout::SaveEnumType(PropertyType propertyType, 
                                        int enumValue, 
                                        DWRITE_TEXT_RANGE textRange)
{
    PropertySetting propertySetting;
    propertySetting.PropertyType = propertyType;
    propertySetting.EnumValue = enumValue;
    propertySetting.TextRange = textRange;

    m_propertySettings.push_back(propertySetting);
}

void SplittableTextLayout::SaveBoolType(PropertyType propertyType, 
                                        BOOL boolValue, 
                                        DWRITE_TEXT_RANGE textRange)
{
    PropertySetting propertySetting;
    propertySetting.PropertyType = propertyType;
    propertySetting.Boolean = boolValue;
    propertySetting.TextRange = textRange;

    m_propertySettings.push_back(propertySetting);
}

HRESULT SplittableTextLayout::TransferPropertySetting(SplittableTextLayout* textLayout, 
                                                      PropertySetting propertySetting, 
                                                      DWRITE_TEXT_RANGE textRange)
{
    HRESULT hr = S_OK;

    switch (propertySetting.PropertyType)
    {
    case PropertyType::DrawingEffect:
        hr = textLayout->SetDrawingEffect(propertySetting.IUnknownPointer.Get(), 
                                          textRange);
        break;

    case PropertyType::FontCollection:
        hr = textLayout->SetFontCollection((IDWriteFontCollection *) 
                                                propertySetting.IUnknownPointer.Get(), 
                                           textRange);
        break;

    case PropertyType::FontFamilyName:
        hr = textLayout->SetFontFamilyName(propertySetting.StringPointer, 
                                           textRange);
        break;

    case PropertyType::FontSize:
        hr = textLayout->SetFontSize(propertySetting.FloatValue[0], 
                                     textRange);
        break;

    case PropertyType::FontStretch:
        hr = textLayout->SetFontStretch((DWRITE_FONT_STRETCH) 
                                            propertySetting.EnumValue, 
                                        textRange);
        break;

    case PropertyType::FontStyle:
        hr = textLayout->SetFontStyle((DWRITE_FONT_STYLE) 
                                            propertySetting.EnumValue, 
                                      textRange);
        break;

    case PropertyType::FontWeight:
        hr = textLayout->SetFontWeight((DWRITE_FONT_WEIGHT) 
                                            propertySetting.EnumValue, 
                                       textRange);
        break;

    case PropertyType::InlineObject:
        hr = textLayout->SetInlineObject((IDWriteInlineObject *) 
                                            propertySetting.IUnknownPointer.Get(), 
                                         textRange);
        break;

    case PropertyType::LocaleName:
        hr = textLayout->SetLocaleName(propertySetting.StringPointer, 
                                       textRange);
        break;

    case PropertyType::Strikethrough:
        hr = textLayout->SetStrikethrough(propertySetting.Boolean, 
                                          textRange);
        break;

    case PropertyType::Typography:
        hr = textLayout->SetTypography((IDWriteTypography *) 
                                            propertySetting.IUnknownPointer.Get(), 
                                       textRange);
        break;

    case PropertyType::Underline:
        hr = textLayout->SetUnderline(propertySetting.Boolean, 
                                      textRange);
        break;

    case PropertyType::CharacterSpacing:
        hr = textLayout->SetCharacterSpacing(propertySetting.FloatValue[0], 
                                             propertySetting.FloatValue[1], 
                                             propertySetting.FloatValue[2], 
                                             textRange);
        break;

    case PropertyType::PairKerning:
        hr = textLayout->SetPairKerning(propertySetting.Boolean, 
                                        textRange);
        break;
    }
    return hr;
}