﻿#include "pch.h"
#include "PaginationExperiment6Renderer.h"
#include "Common/DirectXHelper.h"

using namespace PaginationExperiment6;

using namespace D2D1;
using namespace Microsoft::WRL;

PaginationExperiment6Renderer::PaginationExperiment6Renderer(const std::shared_ptr<DeviceResources>& deviceResources) :
    m_needsRedraw(false),
    m_currentScroll(0),
    m_maximumScroll(0),
	m_deviceResources(deviceResources)
{
	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

    // IDWriteTextFormat for page numbers
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(L"Segoe UI",
                                                                nullptr,
                                                                DWRITE_FONT_WEIGHT_NORMAL,
                                                                DWRITE_FONT_STYLE_NORMAL,
                                                                DWRITE_FONT_STRETCH_NORMAL,
                                                                18.0f,
                                                                L"en-US",
                                                                &m_textFormatPageNumber)
        );

    DX::ThrowIfFailed(
        m_textFormatPageNumber->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
        );

    DX::ThrowIfFailed(
        m_textFormatPageNumber->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_CENTER)
        );

    // IDWriteTextFormat for "Paginate..." status
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(L"Segoe UI",
                                                                nullptr,
                                                                DWRITE_FONT_WEIGHT_NORMAL,
                                                                DWRITE_FONT_STYLE_NORMAL,
                                                                DWRITE_FONT_STRETCH_NORMAL,
                                                                48.0f,
                                                                L"en-US",
                                                                &m_textFormatStatus)
        );

    DX::ThrowIfFailed(
        m_textFormatStatus->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_CENTER)
        );

    DX::ThrowIfFailed(
        m_textFormatStatus->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_CENTER)
        );

	CreateDeviceDependentResources();
}

void PaginationExperiment6Renderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}
void PaginationExperiment6Renderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void PaginationExperiment6Renderer::CreateWindowSizeDependentResources()
{
    PaginateBook();
    m_needsRedraw = true;
}

void PaginationExperiment6Renderer::ClearPages()
{
    m_pages.clear();
    m_needsRedraw = true;
}

void PaginationExperiment6Renderer::SetBookText(std::vector<std::wstring> lines)
{
    ComPtr<IDWriteFactory> dwriteFactory = m_deviceResources->GetDWriteFactory();

    // Create a basic IDWriteTextFormat for all book text
    ComPtr<IDWriteTextFormat> textFormat;

    DX::ThrowIfFailed(
        dwriteFactory->CreateTextFormat(L"Century Schoolbook",
                                        nullptr,
                                        DWRITE_FONT_WEIGHT_NORMAL,
                                        DWRITE_FONT_STYLE_NORMAL,
                                        DWRITE_FONT_STRETCH_NORMAL,
                                        24.0f,
                                        L"en-US",
                                        &textFormat)
        );

    // Create first-line indent object
    ComPtr<IDWriteInlineObject> halfInchIndent = new FirstLineIndent(48);

    // Clear out any earlier paragraphs
    m_paragraphs.clear();

    // Loop through the single-line paragraphs
    for (unsigned int index = 0; index < lines.size(); index++)
    {
        // Set the text with an initial character for first-line indenting
        Paragraph paragraph;
        std::wstring text = L"\x200B" + lines.at(index);

        // Create the SplittableTextLayout object
        DX::ThrowIfFailed(
            SplittableTextLayout::Create(text.c_str(),
                                         text.length(),
                                         textFormat.Get(),
                                         0, 0,
                                         &paragraph.TextLayout)
            );

        if (lines[index][0] != ' ')
        {
            // Justify paragraphs that don't begin with a space
            DX::ThrowIfFailed(
                paragraph.TextLayout->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_JUSTIFIED)
                );

            // And first-line indent by a half inch
            DWRITE_TEXT_RANGE textRange = { 0, 1 };

            DX::ThrowIfFailed(
                paragraph.TextLayout->SetInlineObject(halfInchIndent.Get(), textRange)
                );
        }
        else
        {
            // Set smaller fixed-pitch font for other paragraphs
            DWRITE_TEXT_RANGE textRange;
            textRange.startPosition = 0;
            textRange.length = text.length();

            DX::ThrowIfFailed(
                paragraph.TextLayout->SetFontFamilyName(L"Courier New", textRange)
                );

            DX::ThrowIfFailed(
                paragraph.TextLayout->SetFontSize(18.0f, textRange)
                );
        }

        paragraph.SpaceAfter = 12.0f;
        paragraph.CanAdjustLineSpacing = true;
        m_paragraphs.push_back(paragraph);
    }

    // Now perform pagination
    PaginateBook();
    m_needsRedraw = true;
}

void PaginationExperiment6Renderer::PaginateBook()
{
    if (m_paragraphs.size() == 0)
        return;

    Windows::Foundation::Size windowSize = m_deviceResources->GetLogicalSize();

    if (windowSize.IsEmpty)
        return;

    // Set some constants and calculate page and text sizes
    float pageHeight = windowSize.Height;
    float topMargin = 25;
    float bottomMargin = 40;
    float textHeight = pageHeight - topMargin - bottomMargin;

    float pageWidth = min(650, windowSize.Width);
    float leftMargin = 25;
    float rightMargin = 25;
    float textWidth = pageWidth - leftMargin - rightMargin;

    // Prepare assembling pages
    m_pages.clear();

    // Prepare for the first page
    float x = leftMargin;
    float y = topMargin;
    float maxY = pageHeight - bottomMargin;
    Page page;
    page.Size = SizeF(pageWidth, pageHeight);   // uniform page size

    // Loop through all the paragraphs
    for (size_t i = 0; i < m_paragraphs.size(); i++)
    {
        Paragraph paragraph = m_paragraphs.at(i);

        // Set paragraph height and width
        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxWidth(textWidth)
            );

        DX::ThrowIfFailed(
            paragraph.TextLayout->SetMaxHeight(FloatMax())
            );

        // Add paragraph to page (or possibly pages)
        bool isSplit = false;

        do
        {
            while (AddParagraphToPage(page, paragraph, x, y, maxY, isSplit))
            {
                // End of page so adjust line spacing & add page to collection
                MakePageUniformHeight(page, maxY);
                m_pages.push_back(page);

                // Begin a new page
                x = leftMargin;
                y = topMargin;
                page.Paragraphs.clear();
            }
        } while (isSplit);
    }

    // Don't adjust line spacing on this last page!
    m_pages.push_back(page);

    // Calculate the maximum scroll
    m_maximumScroll = m_pages.size() * pageWidth - windowSize.Width;
    ScrollHorizontalDelta(0);
}

bool PaginationExperiment6Renderer::AddParagraphToPage(Page& page, 
                                                       Paragraph& paragraph, 
                                                       float x, 
                                                       float& y, 
                                                       float maxY, 
                                                       bool& isSplit)
{
    // We already know where the paragraph is going
    paragraph.Origin = Point2F(x, y);

    // Obtain the height of the paragraph
    DWRITE_TEXT_METRICS textMetrics;

    DX::ThrowIfFailed(
        paragraph.TextLayout->GetMetrics(&textMetrics)
        );

    // Save information in Paragraph object
    paragraph.LineCount = textMetrics.lineCount;
    paragraph.Height = textMetrics.height;

    // Check if it overruns the maximum height
    if (y + paragraph.Height > maxY)
    {
        // Paragraph is too long, check for a split
        UINT32 lineCount = textMetrics.lineCount;

        // We already know the paragraph is too long,
        //  so if it consists of only one line, it won't fit.
        // If it has two lines, only one line will fit,
        //  and that creates both a widow and orphan.
        // If it has three lines, either an orphan or 
        //  widow will result.
        // Hence...
        if (paragraph.LineCount <= 3)
        {
            isSplit = false;
            return true;            // end of page
        }

        // Get the line metrics
        std::vector<DWRITE_LINE_METRICS> lineMetrics(lineCount);

        DX::ThrowIfFailed(
            paragraph.TextLayout->GetLineMetrics(lineMetrics.data(), 
                                                 lineMetrics.size(), 
                                                 &lineCount)
            );

        // Accumulate the number of lines and character count that fit
        int fittableLineCount = 0;
        float fittableHeight = 0;
        int characterCount = 0;

        for (DWRITE_LINE_METRICS lineMetric : lineMetrics)
        {
            // Check for next line that exceeds page height
            if (y + lineMetric.height > maxY)
            {
                break;
            }

            fittableLineCount++;
            fittableHeight += lineMetric.height;
            characterCount += lineMetric.length;
            y += lineMetric.height;

            // Check for maximum lines before a widow is made
            if (fittableLineCount + 2 == lineCount)
            {
                break;
            }
        }
        
        // Check for 0 or 1 line (an orphan) that fits on the page
        if (fittableLineCount <= 1)
        {
            isSplit = false;
            return true;            // end of page
        }
        
        // Ready for a valid split!
        ComPtr<SplittableTextLayout> textLayout1;
        ComPtr<SplittableTextLayout> textLayout2;

        DX::ThrowIfFailed(
            paragraph.TextLayout->Split(characterCount, &textLayout1, &textLayout2)
            );

        // This is the displayable first half of the paragraph.
        // Create a new Paragraph with this new layout.
        Paragraph newParagraph = paragraph;
        newParagraph.TextLayout = textLayout1;
        newParagraph.Height = fittableHeight;
        newParagraph.SpaceAfter = 0;            // because at end of page
        newParagraph.LineCount = fittableLineCount;
        page.Paragraphs.push_back(newParagraph);
        page.TextBottom = y;

        // This is the second half of the paragraph,
        //  to be handled in next call to method.
        paragraph.TextLayout = textLayout2;

        isSplit = true;
        return true;            // end of page
    }

    y += paragraph.Height + paragraph.SpaceAfter;

    // Create a new Paragraph with a cloned SplittableTextLayout
    Paragraph newParagraph = paragraph;

    DX::ThrowIfFailed(
        paragraph.TextLayout->Clone(&newParagraph.TextLayout)
        );

    // Add it to the page and update page height
    page.Paragraphs.push_back(newParagraph);
    page.TextBottom = y - paragraph.SpaceAfter;
    isSplit = false;
    return false;               // not end of page
}

void PaginationExperiment6Renderer::MakePageUniformHeight(Page& page, 
                                                          float desiredHeight)
{
    int totalLineCount = 0;

    for (Paragraph& paragraph : page.Paragraphs)
    {
        if (paragraph.CanAdjustLineSpacing)
        {
            totalLineCount += paragraph.LineCount;
        }
    }

    float extraLineSpacing = (desiredHeight - page.TextBottom) / totalLineCount;
    D2D1_POINT_2F origin = page.Paragraphs[0].Origin;

    for (Paragraph& paragraph : page.Paragraphs)
    {
        // Set new origin
        paragraph.Origin = origin;

        if (paragraph.CanAdjustLineSpacing)
        {
            // Calculate and set new line spacing
            float lineSpacing = paragraph.Height / paragraph.LineCount;
            lineSpacing += extraLineSpacing;

            DX::ThrowIfFailed(
                paragraph.TextLayout->SetLineSpacing
                                (DWRITE_LINE_SPACING_METHOD_UNIFORM,
                                 lineSpacing, 0.8f * lineSpacing)
                );

            // Set new paragraph height
            paragraph.Height = lineSpacing * paragraph.LineCount;
        }

        // Calculate origin of next paragraph
        origin.y += paragraph.Height + paragraph.SpaceAfter;
    }
}

void PaginationExperiment6Renderer::ScrollHorizontalDelta(float deltaX)
{
    m_currentScroll = max(0, min(m_maximumScroll, m_currentScroll - deltaX));
    m_needsRedraw = true;
}

void PaginationExperiment6Renderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
bool PaginationExperiment6Renderer::Render()
{
    if (!m_needsRedraw)
        return false;

	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::White));

    // Display status 
    if (m_pages.size() == 0)
    {
        std::wstring text = L"Paginating...";
        D2D1_RECT_F layoutRect = RectF(0, 0, logicalSize.Width, 
                                             logicalSize.Height);
        context->DrawText(text.c_str(), 
                          text.length(), 
                          m_textFormatStatus.Get(), 
                          &layoutRect, 
                          m_blackBrush.Get());
    }

    float x = -m_currentScroll;
    int pageNumber = 0;

    //Loop through pages
    for (const Page& page : m_pages)
    {
        pageNumber++;

        if (x + page.Size.width > 0 && x < logicalSize.Width)
        {
            context->SetTransform(Matrix3x2F::Translation(x, 0) *
                        m_deviceResources->GetOrientationTransform2D());

            for (const Paragraph& paragraph : page.Paragraphs)
            {
                context->DrawTextLayout(paragraph.Origin,
                    paragraph.TextLayout.Get(),
                    m_blackBrush.Get());

                // Set page number
                std::wstring text = L"Page " + std::to_wstring(pageNumber) + 
                                    L" of " + std::to_wstring(m_pages.size());

                D2D1_RECT_F layoutRect = RectF(0, 0, page.Size.width, 
                                                     page.Size.height);
                context->DrawText(text.c_str(), 
                                  text.length(), 
                                  m_textFormatPageNumber.Get(), 
                                  &layoutRect, 
                                  m_blackBrush.Get());
            }
        }

        x += page.Size.width;
    }

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());

    m_needsRedraw = false;
    return true;
}

