﻿#include "pch.h"
#include "PaginationExperiment6Main.h"
#include "Common\DirectXHelper.h"

using namespace PaginationExperiment6;
using namespace Windows::Foundation;
using namespace Windows::System::Threading;
using namespace Concurrency;

// Loads and initializes application assets when the application is loaded.
PaginationExperiment6Main::PaginationExperiment6Main(const std::shared_ptr<DeviceResources>& deviceResources) :
	m_deviceResources(deviceResources)
{
	// Register to be notified if the Device is lost or recreated
	m_deviceResources->RegisterDeviceNotify(this);

	// Content initialization.
	m_PaginationExperiment6Renderer = std::unique_ptr<PaginationExperiment6Renderer>(new PaginationExperiment6Renderer(m_deviceResources));

	// TODO: Change the timer settings if you want something other than the default variable timestep mode.
	// e.g. for 60 FPS fixed timestep update logic, call:
	/*
	m_timer.SetFixedTimeStep(true);
	m_timer.SetTargetElapsedSeconds(1.0 / 60);
	*/
}

PaginationExperiment6Main::~PaginationExperiment6Main()
{
	// Deregister device notification
	m_deviceResources->RegisterDeviceNotify(nullptr);
}

// Updates application state when the window size changes (e.g. device orientation change)
void PaginationExperiment6Main::CreateWindowSizeDependentResources() 
{
    m_PaginationExperiment6Renderer->ClearPages();

    concurrency::create_async([this]()
    {
        critical_section::scoped_lock lock(m_criticalSection);
        m_PaginationExperiment6Renderer->CreateWindowSizeDependentResources();
    });
}

void PaginationExperiment6Main::StartRenderLoop()
{
	// If the animation render loop is already running then do not start another thread.
	if (m_renderLoopWorker != nullptr && m_renderLoopWorker->Status == AsyncStatus::Started)
	{
		return;
	}

	// Create a task that will be run on a background thread.
	auto workItemHandler = ref new WorkItemHandler([this](IAsyncAction ^ action)
	{
		// Calculate the updated frame and render once per vertical blanking interval.
		while (action->Status == AsyncStatus::Started)
		{
			critical_section::scoped_lock lock(m_criticalSection);
			Update();
			if (Render())
			{
				m_deviceResources->Present();
			}
		}
	});

	// Run task on a dedicated high priority background thread.
	m_renderLoopWorker = ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);
}

void PaginationExperiment6Main::StopRenderLoop()
{
	m_renderLoopWorker->Cancel();
}

// Updates the application state once per frame.
void PaginationExperiment6Main::Update() 
{
	ProcessInput();

	// Update scene objects.
	m_timer.Tick([&]()
	{
		// Content update functions.
		m_PaginationExperiment6Renderer->Update(m_timer);
	});
}

// Process all input from the user before updating game state
void PaginationExperiment6Main::ProcessInput()
{
	// TODO: Add per frame input handling here.
}

// Renders the current frame according to the current application state.
// Returns true if the frame was rendered and is ready to be displayed.
bool PaginationExperiment6Main::Render() 
{
	// Don't try to render anything before the first Update.
	if (m_timer.GetFrameCount() == 0)
	{
		return false;
	}

	// Render the screen.
    return m_PaginationExperiment6Renderer->Render();

}

// Notifies renderers that device resources need to be released.
void PaginationExperiment6Main::OnDeviceLost()
{
	m_PaginationExperiment6Renderer->ReleaseDeviceDependentResources();
}

// Notifies renderers that device resources may now be recreated.
void PaginationExperiment6Main::OnDeviceRestored()
{
	m_PaginationExperiment6Renderer->CreateDeviceDependentResources();
	CreateWindowSizeDependentResources();
}

// Methods called from DirectXPage
void PaginationExperiment6Main::SetBookText(std::vector<std::wstring> lines)
{
    m_PaginationExperiment6Renderer->SetBookText(lines);
}

void PaginationExperiment6Main::ScrollHorizontalDelta(float deltaX)
{
    m_PaginationExperiment6Renderer->ScrollHorizontalDelta(deltaX);
}

